# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

filelist = {
    "real": {
        "2s1": [
            "2s1_real_A_elevDeg_015_azCenter_010_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_010_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_011_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_011_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_012_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_012_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_013_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_013_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_014_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_014_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_015_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_015_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_016_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_016_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_017_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_017_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_018_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_018_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_019_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_019_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_020_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_020_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_021_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_021_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_022_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_022_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_023_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_023_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_024_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_024_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_025_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_025_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_026_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_026_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_027_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_027_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_028_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_028_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_029_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_029_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_030_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_030_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_031_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_031_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_032_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_032_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_033_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_033_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_034_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_034_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_035_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_035_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_036_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_036_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_038_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_038_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_039_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_039_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_040_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_040_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_041_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_041_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_042_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_042_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_043_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_043_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_044_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_044_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_045_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_045_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_046_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_046_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_048_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_048_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_049_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_049_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_050_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_050_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_051_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_051_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_052_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_052_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_053_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_053_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_054_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_054_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_055_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_055_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_056_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_056_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_057_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_057_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_058_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_058_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_059_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_059_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_060_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_060_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_061_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_061_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_062_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_062_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_063_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_063_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_064_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_064_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_066_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_066_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_068_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_068_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_069_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_069_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_070_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_070_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_071_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_071_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_072_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_072_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_073_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_073_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_074_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_074_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_075_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_075_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_076_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_076_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_077_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_077_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_078_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_078_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_079_22_serial_b01.mat",
            "2s1_real_A_elevDeg_015_azCenter_079_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_010_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_010_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_011_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_011_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_012_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_012_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_013_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_013_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_014_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_014_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_015_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_015_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_016_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_016_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_017_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_017_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_018_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_018_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_019_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_019_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_020_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_020_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_021_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_021_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_023_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_023_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_024_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_024_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_025_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_025_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_026_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_026_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_027_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_027_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_029_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_029_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_030_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_030_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_031_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_031_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_032_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_032_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_033_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_033_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_034_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_034_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_035_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_035_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_036_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_036_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_037_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_037_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_038_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_038_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_039_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_039_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_041_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_041_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_042_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_042_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_044_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_044_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_046_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_046_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_047_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_047_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_049_22_serial_b01.mat",
            "2s1_real_A_elevDeg_016_azCenter_049_22_serial_b01.mat",
        ],
        "bmp2": [
            "bmp2_real_A_elevDeg_016_azCenter_014_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_014_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_016_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_016_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_017_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_017_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_018_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_018_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_021_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_021_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_022_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_022_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_023_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_023_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_028_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_028_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_029_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_029_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_030_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_030_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_031_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_031_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_033_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_033_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_034_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_034_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_035_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_035_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_036_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_036_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_037_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_037_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_038_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_038_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_039_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_039_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_040_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_040_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_041_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_041_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_042_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_042_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_043_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_043_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_044_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_044_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_045_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_045_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_047_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_047_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_048_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_048_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_049_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_049_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_050_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_050_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_051_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_051_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_052_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_052_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_053_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_053_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_054_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_054_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_055_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_055_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_056_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_056_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_057_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_057_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_058_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_058_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_059_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_059_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_060_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_060_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_061_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_061_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_062_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_062_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_063_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_063_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_064_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_064_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_065_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_065_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_066_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_066_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_067_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_067_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_069_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_069_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_070_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_070_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_071_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_071_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_072_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_072_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_073_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_073_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_074_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_074_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_075_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_075_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_076_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_076_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_077_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_077_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_079_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_016_azCenter_079_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_012_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_012_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_013_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_013_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_014_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_014_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_015_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_015_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_017_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_017_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_018_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_018_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_019_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_019_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_020_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_020_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_023_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_023_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_024_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_024_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_025_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_025_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_026_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_026_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_027_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_027_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_028_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_028_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_029_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_029_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_030_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_030_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_032_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_032_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_033_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_033_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_034_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_034_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_035_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_035_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_036_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_036_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_037_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_037_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_038_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_038_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_039_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_039_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_042_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_042_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_043_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_043_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_044_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_044_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_045_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_045_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_046_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_046_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_047_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_047_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_048_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_048_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_049_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_049_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_050_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_050_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_051_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_051_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_054_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_054_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_055_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_055_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_058_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_058_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_059_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_059_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_060_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_060_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_062_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_062_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_063_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_063_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_064_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_064_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_065_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_065_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_067_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_067_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_068_49_serial_9563.mat",
            "bmp2_real_A_elevDeg_017_azCenter_068_49_serial_9563.mat",
        ],
        "btr70": [
            "btr70_real_A_elevDeg_016_azCenter_011_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_011_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_012_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_012_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_013_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_013_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_014_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_014_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_015_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_015_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_016_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_016_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_017_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_017_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_018_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_018_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_019_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_019_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_020_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_020_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_021_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_021_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_022_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_022_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_023_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_023_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_025_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_025_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_026_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_026_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_027_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_027_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_028_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_028_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_029_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_029_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_030_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_030_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_031_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_031_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_032_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_032_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_033_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_033_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_035_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_035_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_036_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_036_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_037_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_037_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_038_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_038_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_041_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_041_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_044_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_044_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_045_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_045_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_046_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_046_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_050_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_050_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_058_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_058_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_060_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_060_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_062_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_062_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_063_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_063_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_065_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_065_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_066_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_066_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_067_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_067_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_068_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_068_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_070_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_070_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_072_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_072_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_075_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_075_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_079_00_serial_c71.mat",
            "btr70_real_A_elevDeg_016_azCenter_079_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_011_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_011_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_014_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_014_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_015_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_015_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_016_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_016_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_018_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_018_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_019_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_019_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_020_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_020_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_021_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_021_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_023_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_023_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_024_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_024_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_025_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_025_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_026_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_026_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_029_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_029_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_030_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_030_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_032_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_032_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_034_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_034_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_035_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_035_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_037_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_037_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_038_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_038_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_039_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_039_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_040_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_040_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_043_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_043_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_044_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_044_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_045_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_045_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_046_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_046_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_047_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_047_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_048_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_048_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_049_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_049_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_050_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_050_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_051_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_051_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_052_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_052_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_053_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_053_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_054_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_054_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_056_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_056_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_057_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_057_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_058_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_058_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_061_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_061_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_062_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_062_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_063_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_063_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_064_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_064_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_066_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_066_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_067_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_067_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_068_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_068_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_069_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_069_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_071_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_071_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_072_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_072_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_074_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_074_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_075_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_075_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_079_00_serial_c71.mat",
            "btr70_real_A_elevDeg_017_azCenter_079_00_serial_c71.mat",
        ],
        "m1": [
            "m1_real_A_elevDeg_014_azCenter_010_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_010_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_011_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_011_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_012_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_012_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_014_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_014_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_015_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_015_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_016_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_016_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_017_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_017_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_018_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_018_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_019_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_019_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_020_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_020_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_021_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_021_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_022_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_022_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_024_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_024_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_026_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_026_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_028_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_028_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_029_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_029_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_030_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_030_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_031_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_031_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_032_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_032_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_033_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_033_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_034_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_034_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_036_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_036_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_037_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_037_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_038_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_038_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_039_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_039_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_040_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_014_azCenter_040_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_010_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_010_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_011_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_011_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_012_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_012_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_013_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_013_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_014_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_014_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_015_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_015_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_016_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_016_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_017_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_017_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_018_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_018_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_019_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_019_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_020_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_020_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_021_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_021_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_022_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_022_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_023_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_023_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_024_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_024_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_025_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_025_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_026_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_026_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_027_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_027_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_028_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_028_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_029_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_029_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_030_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_030_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_031_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_031_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_032_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_032_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_033_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_033_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_034_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_034_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_035_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_035_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_036_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_036_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_037_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_037_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_038_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_038_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_039_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_039_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_040_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_040_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_041_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_041_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_042_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_042_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_043_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_043_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_044_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_044_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_045_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_045_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_046_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_046_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_047_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_047_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_049_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_049_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_050_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_050_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_051_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_051_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_052_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_052_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_055_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_055_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_058_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_058_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_059_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_059_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_060_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_060_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_064_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_064_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_072_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_072_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_074_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_074_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_076_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_076_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_077_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_077_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_079_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_016_azCenter_079_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_012_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_012_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_013_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_013_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_014_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_014_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_015_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_015_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_016_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_016_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_017_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_017_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_018_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_018_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_019_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_019_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_020_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_020_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_021_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_021_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_024_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_024_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_025_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_025_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_028_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_028_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_029_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_029_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_030_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_030_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_032_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_032_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_033_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_033_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_034_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_034_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_035_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_035_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_037_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_037_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_038_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_038_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_039_18_serial_0ap00n.mat",
            "m1_real_A_elevDeg_017_azCenter_039_18_serial_0ap00n.mat",
        ],
        "m2": [
            "m2_real_A_elevDeg_014_azCenter_011_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_011_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_012_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_012_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_013_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_013_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_014_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_014_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_015_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_015_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_016_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_016_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_017_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_017_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_018_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_018_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_019_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_019_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_021_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_021_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_023_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_023_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_025_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_025_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_026_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_026_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_027_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_027_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_028_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_028_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_029_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_029_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_030_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_030_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_031_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_031_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_033_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_033_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_034_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_034_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_035_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_035_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_036_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_036_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_037_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_014_azCenter_037_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_010_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_010_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_011_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_011_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_012_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_012_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_013_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_013_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_014_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_014_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_015_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_015_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_016_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_016_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_017_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_017_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_018_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_018_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_019_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_019_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_020_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_020_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_021_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_021_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_022_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_022_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_023_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_023_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_024_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_024_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_025_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_025_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_026_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_026_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_027_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_027_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_028_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_028_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_029_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_029_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_030_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_030_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_031_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_031_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_032_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_032_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_033_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_033_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_034_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_034_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_035_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_035_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_036_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_036_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_037_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_037_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_038_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_038_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_039_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_039_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_040_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_040_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_041_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_041_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_042_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_042_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_043_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_043_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_044_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_044_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_046_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_046_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_047_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_047_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_048_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_048_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_049_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_049_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_052_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_052_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_055_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_055_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_056_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_056_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_057_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_057_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_061_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_061_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_069_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_069_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_071_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_071_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_073_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_073_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_074_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_074_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_076_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_076_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_077_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_077_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_078_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_078_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_079_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_016_azCenter_079_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_010_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_010_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_011_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_011_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_012_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_012_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_013_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_013_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_014_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_014_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_015_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_015_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_016_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_016_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_017_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_017_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_018_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_018_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_021_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_021_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_022_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_022_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_025_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_025_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_026_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_026_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_027_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_027_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_029_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_029_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_030_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_030_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_031_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_031_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_032_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_032_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_034_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_034_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_035_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_035_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_036_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_036_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_037_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_037_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_040_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_040_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_041_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_041_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_043_91_serial_mv02gx.mat",
            "m2_real_A_elevDeg_017_azCenter_043_91_serial_mv02gx.mat",
        ],
        "m35": [
            "m35_real_A_elevDeg_014_azCenter_010_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_010_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_012_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_012_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_013_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_013_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_014_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_014_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_015_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_015_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_016_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_016_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_017_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_017_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_018_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_018_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_019_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_019_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_020_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_020_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_022_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_022_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_024_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_024_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_026_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_026_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_027_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_027_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_028_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_028_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_029_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_029_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_030_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_030_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_031_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_031_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_032_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_032_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_034_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_034_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_035_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_035_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_036_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_036_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_037_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_037_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_038_62_serial_t839.mat",
            "m35_real_A_elevDeg_014_azCenter_038_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_010_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_010_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_011_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_011_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_012_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_012_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_013_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_013_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_014_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_014_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_015_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_015_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_016_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_016_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_017_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_017_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_018_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_018_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_019_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_019_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_020_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_020_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_021_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_021_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_022_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_022_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_023_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_023_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_024_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_024_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_025_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_025_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_026_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_026_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_027_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_027_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_028_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_028_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_029_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_029_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_030_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_030_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_031_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_031_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_032_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_032_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_033_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_033_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_034_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_034_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_035_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_035_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_036_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_036_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_037_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_037_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_038_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_038_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_039_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_039_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_040_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_040_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_041_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_041_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_042_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_042_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_043_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_043_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_044_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_044_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_045_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_045_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_047_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_047_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_048_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_048_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_049_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_049_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_050_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_050_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_053_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_053_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_056_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_056_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_057_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_057_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_058_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_058_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_062_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_062_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_070_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_070_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_072_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_072_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_074_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_074_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_075_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_075_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_077_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_077_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_078_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_078_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_079_62_serial_t839.mat",
            "m35_real_A_elevDeg_016_azCenter_079_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_010_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_010_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_011_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_011_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_012_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_012_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_013_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_013_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_014_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_014_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_015_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_015_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_016_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_016_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_017_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_017_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_018_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_018_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_019_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_019_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_022_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_022_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_023_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_023_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_026_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_026_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_027_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_027_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_028_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_028_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_030_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_030_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_031_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_031_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_032_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_032_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_033_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_033_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_035_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_035_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_036_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_036_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_037_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_037_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_038_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_038_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_041_62_serial_t839.mat",
            "m35_real_A_elevDeg_017_azCenter_041_62_serial_t839.mat",
        ],
        "m548": [
            "m548_real_A_elevDeg_014_azCenter_011_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_011_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_012_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_012_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_013_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_013_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_014_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_014_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_015_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_015_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_016_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_016_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_017_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_017_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_018_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_018_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_019_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_019_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_021_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_021_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_023_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_023_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_025_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_025_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_026_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_026_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_027_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_027_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_028_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_028_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_029_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_029_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_030_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_030_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_031_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_031_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_033_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_033_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_034_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_034_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_035_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_035_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_036_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_036_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_037_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_014_azCenter_037_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_010_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_010_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_011_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_011_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_012_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_012_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_013_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_013_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_014_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_014_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_015_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_015_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_016_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_016_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_017_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_017_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_018_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_018_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_019_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_019_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_020_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_020_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_021_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_021_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_022_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_022_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_023_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_023_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_024_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_024_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_025_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_025_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_026_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_026_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_027_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_027_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_028_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_028_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_029_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_029_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_030_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_030_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_031_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_031_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_032_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_032_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_033_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_033_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_034_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_034_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_035_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_035_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_036_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_036_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_037_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_037_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_038_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_038_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_039_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_039_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_040_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_040_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_041_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_041_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_042_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_042_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_043_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_043_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_044_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_044_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_046_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_046_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_047_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_047_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_048_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_048_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_049_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_049_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_052_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_052_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_055_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_055_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_056_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_056_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_057_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_057_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_061_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_061_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_069_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_069_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_071_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_071_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_073_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_073_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_074_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_074_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_076_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_076_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_077_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_077_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_078_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_078_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_079_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_016_azCenter_079_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_010_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_010_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_011_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_011_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_012_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_012_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_013_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_013_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_014_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_014_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_015_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_015_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_016_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_016_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_017_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_017_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_018_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_018_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_021_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_021_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_022_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_022_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_025_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_025_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_026_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_026_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_027_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_027_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_029_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_029_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_030_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_030_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_031_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_031_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_032_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_032_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_034_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_034_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_035_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_035_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_036_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_036_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_037_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_037_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_040_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_040_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_041_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_041_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_043_63_serial_c245hab.mat",
            "m548_real_A_elevDeg_017_azCenter_043_63_serial_c245hab.mat",
        ],
        "m60": [
            "m60_real_A_elevDeg_015_azCenter_010_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_010_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_011_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_011_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_012_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_012_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_013_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_013_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_014_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_014_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_015_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_015_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_016_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_016_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_017_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_017_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_018_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_018_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_019_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_019_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_020_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_020_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_021_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_021_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_022_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_022_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_023_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_023_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_024_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_024_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_025_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_025_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_026_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_026_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_027_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_027_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_028_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_028_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_029_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_029_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_030_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_030_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_031_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_031_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_032_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_032_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_033_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_033_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_034_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_034_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_035_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_035_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_036_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_036_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_037_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_037_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_038_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_038_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_039_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_039_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_040_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_040_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_042_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_042_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_043_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_043_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_044_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_044_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_045_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_045_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_046_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_046_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_047_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_047_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_048_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_048_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_049_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_049_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_050_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_050_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_052_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_052_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_053_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_053_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_054_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_054_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_055_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_055_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_056_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_056_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_057_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_057_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_058_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_058_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_059_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_059_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_060_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_060_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_061_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_061_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_062_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_062_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_063_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_063_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_064_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_064_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_065_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_065_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_066_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_066_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_067_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_067_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_068_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_068_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_070_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_070_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_073_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_073_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_074_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_074_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_075_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_075_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_076_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_076_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_077_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_077_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_078_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_078_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_079_74_serial_3336.mat",
            "m60_real_A_elevDeg_015_azCenter_079_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_010_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_010_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_011_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_011_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_012_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_012_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_013_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_013_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_014_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_014_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_015_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_015_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_016_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_016_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_017_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_017_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_018_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_018_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_019_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_019_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_020_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_020_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_021_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_021_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_022_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_022_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_023_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_023_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_024_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_024_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_025_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_025_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_027_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_027_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_028_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_028_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_029_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_029_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_030_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_030_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_031_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_031_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_033_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_033_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_034_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_034_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_035_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_035_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_036_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_036_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_037_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_037_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_038_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_038_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_039_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_039_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_040_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_040_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_041_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_041_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_042_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_042_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_043_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_043_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_045_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_045_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_046_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_046_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_048_74_serial_3336.mat",
            "m60_real_A_elevDeg_016_azCenter_048_74_serial_3336.mat",
        ],
        "t72": [
            "t72_real_A_elevDeg_016_azCenter_013_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_013_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_015_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_015_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_016_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_016_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_017_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_017_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_020_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_020_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_021_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_021_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_022_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_022_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_027_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_027_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_028_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_028_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_029_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_029_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_030_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_030_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_032_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_032_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_033_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_033_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_034_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_034_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_035_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_035_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_036_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_036_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_037_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_037_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_038_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_038_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_039_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_039_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_040_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_040_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_041_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_041_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_042_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_042_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_043_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_043_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_044_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_044_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_046_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_046_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_047_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_047_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_048_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_048_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_049_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_049_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_050_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_050_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_051_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_051_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_052_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_052_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_053_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_053_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_054_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_054_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_055_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_055_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_056_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_056_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_057_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_057_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_058_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_058_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_059_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_059_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_060_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_060_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_061_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_061_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_062_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_062_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_063_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_063_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_064_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_064_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_065_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_065_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_066_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_066_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_068_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_068_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_069_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_069_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_070_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_070_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_071_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_071_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_072_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_072_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_073_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_073_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_074_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_074_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_075_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_075_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_076_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_076_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_078_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_078_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_079_77_serial_812.mat",
            "t72_real_A_elevDeg_016_azCenter_079_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_011_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_011_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_012_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_012_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_013_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_013_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_014_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_014_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_016_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_016_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_017_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_017_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_018_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_018_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_019_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_019_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_022_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_022_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_023_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_023_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_024_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_024_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_025_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_025_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_026_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_026_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_027_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_027_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_028_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_028_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_029_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_029_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_031_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_031_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_032_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_032_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_033_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_033_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_034_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_034_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_035_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_035_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_036_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_036_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_037_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_037_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_038_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_038_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_041_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_041_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_042_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_042_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_043_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_043_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_044_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_044_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_045_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_045_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_046_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_046_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_047_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_047_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_048_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_048_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_049_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_049_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_050_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_050_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_053_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_053_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_054_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_054_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_057_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_057_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_058_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_058_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_059_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_059_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_061_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_061_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_062_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_062_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_063_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_063_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_064_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_064_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_066_77_serial_812.mat",
            "t72_real_A_elevDeg_017_azCenter_066_77_serial_812.mat",
        ],
        "zsu23": [
            "zsu23_real_A_elevDeg_015_azCenter_010_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_010_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_011_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_011_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_012_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_012_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_013_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_013_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_014_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_014_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_015_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_015_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_016_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_016_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_017_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_017_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_018_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_018_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_019_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_019_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_020_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_020_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_021_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_021_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_022_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_022_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_023_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_023_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_024_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_024_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_025_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_025_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_026_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_026_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_027_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_027_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_028_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_028_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_029_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_029_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_030_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_030_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_031_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_031_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_032_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_032_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_033_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_033_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_034_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_034_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_035_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_035_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_036_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_036_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_037_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_037_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_039_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_039_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_040_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_040_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_041_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_041_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_042_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_042_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_043_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_043_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_044_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_044_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_045_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_045_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_046_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_046_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_047_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_047_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_049_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_049_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_050_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_050_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_051_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_051_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_052_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_052_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_053_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_053_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_054_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_054_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_055_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_055_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_056_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_056_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_057_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_057_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_058_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_058_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_059_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_059_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_060_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_060_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_061_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_061_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_062_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_062_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_063_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_063_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_064_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_064_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_065_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_065_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_067_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_067_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_069_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_069_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_070_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_070_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_071_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_071_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_072_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_072_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_073_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_073_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_074_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_074_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_075_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_075_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_076_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_076_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_077_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_077_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_078_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_078_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_079_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_015_azCenter_079_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_010_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_010_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_011_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_011_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_012_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_012_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_013_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_013_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_014_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_014_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_015_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_015_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_016_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_016_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_017_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_017_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_018_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_018_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_019_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_019_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_020_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_020_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_021_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_021_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_022_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_022_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_024_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_024_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_025_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_025_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_026_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_026_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_027_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_027_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_028_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_028_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_030_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_030_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_031_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_031_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_032_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_032_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_033_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_033_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_034_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_034_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_035_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_035_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_036_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_036_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_037_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_037_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_038_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_038_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_039_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_039_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_040_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_040_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_042_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_042_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_043_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_043_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_045_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_045_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_047_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_047_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_048_99_serial_d08.mat",
            "zsu23_real_A_elevDeg_016_azCenter_048_99_serial_d08.mat",
        ],
    },
    "synth": {
        "2s1": [
            "2s1_synth_A_elevDeg_015_azCenter_010_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_010_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_011_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_011_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_012_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_012_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_013_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_013_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_014_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_014_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_015_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_015_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_016_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_016_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_017_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_017_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_018_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_018_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_019_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_019_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_020_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_020_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_021_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_021_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_022_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_022_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_023_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_023_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_024_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_024_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_025_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_025_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_026_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_026_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_027_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_027_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_028_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_028_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_029_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_029_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_030_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_030_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_031_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_031_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_032_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_032_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_033_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_033_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_034_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_034_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_035_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_035_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_036_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_036_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_038_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_038_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_039_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_039_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_040_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_040_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_041_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_041_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_042_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_042_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_043_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_043_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_044_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_044_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_045_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_045_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_046_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_046_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_048_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_048_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_049_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_049_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_050_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_050_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_051_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_051_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_052_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_052_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_053_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_053_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_054_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_054_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_055_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_055_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_056_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_056_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_057_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_057_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_058_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_058_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_059_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_059_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_060_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_060_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_061_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_061_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_062_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_062_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_063_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_063_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_064_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_064_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_066_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_066_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_068_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_068_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_069_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_069_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_070_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_070_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_071_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_071_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_072_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_072_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_073_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_073_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_074_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_074_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_075_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_075_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_076_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_076_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_077_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_077_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_078_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_078_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_079_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_015_azCenter_079_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_010_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_010_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_011_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_011_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_012_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_012_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_013_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_013_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_014_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_014_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_015_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_015_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_016_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_016_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_017_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_017_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_018_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_018_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_019_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_019_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_020_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_020_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_021_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_021_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_023_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_023_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_024_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_024_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_025_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_025_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_026_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_026_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_027_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_027_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_029_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_029_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_030_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_030_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_031_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_031_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_032_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_032_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_033_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_033_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_034_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_034_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_035_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_035_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_036_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_036_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_037_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_037_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_038_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_038_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_039_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_039_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_041_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_041_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_042_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_042_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_044_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_044_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_046_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_046_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_047_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_047_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_049_22_serial_b01.mat",
            "2s1_synth_A_elevDeg_016_azCenter_049_22_serial_b01.mat",
        ],
        "bmp2": [
            "bmp2_synth_A_elevDeg_016_azCenter_014_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_014_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_016_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_016_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_017_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_017_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_018_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_018_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_021_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_021_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_022_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_022_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_023_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_023_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_028_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_028_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_029_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_029_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_030_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_030_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_031_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_031_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_033_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_033_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_034_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_034_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_035_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_035_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_036_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_036_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_037_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_037_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_038_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_038_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_039_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_039_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_040_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_040_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_041_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_041_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_042_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_042_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_043_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_043_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_044_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_044_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_045_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_045_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_047_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_047_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_048_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_048_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_049_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_049_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_050_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_050_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_051_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_051_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_052_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_052_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_053_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_053_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_054_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_054_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_055_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_055_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_056_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_056_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_057_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_057_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_058_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_058_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_059_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_059_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_060_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_060_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_061_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_061_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_062_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_062_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_063_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_063_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_064_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_064_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_065_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_065_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_066_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_066_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_067_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_067_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_069_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_069_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_070_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_070_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_071_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_071_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_072_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_072_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_073_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_073_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_074_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_074_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_075_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_075_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_076_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_076_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_077_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_077_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_079_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_016_azCenter_079_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_012_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_012_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_013_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_013_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_014_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_014_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_015_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_015_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_017_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_017_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_018_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_018_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_019_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_019_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_020_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_020_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_023_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_023_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_024_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_024_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_025_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_025_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_026_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_026_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_027_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_027_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_028_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_028_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_029_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_029_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_030_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_030_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_032_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_032_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_033_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_033_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_034_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_034_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_035_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_035_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_036_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_036_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_037_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_037_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_038_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_038_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_039_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_039_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_042_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_042_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_043_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_043_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_044_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_044_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_045_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_045_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_046_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_046_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_047_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_047_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_048_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_048_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_049_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_049_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_050_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_050_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_051_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_051_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_054_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_054_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_055_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_055_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_058_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_058_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_059_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_059_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_060_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_060_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_062_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_062_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_063_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_063_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_064_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_064_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_065_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_065_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_067_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_067_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_068_49_serial_9563.mat",
            "bmp2_synth_A_elevDeg_017_azCenter_068_49_serial_9563.mat",
        ],
        "btr70": [
            "btr70_synth_A_elevDeg_016_azCenter_011_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_011_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_012_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_012_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_013_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_013_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_014_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_014_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_015_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_015_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_016_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_016_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_017_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_017_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_018_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_018_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_019_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_019_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_020_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_020_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_021_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_021_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_022_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_022_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_023_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_023_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_025_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_025_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_026_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_026_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_027_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_027_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_028_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_028_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_029_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_029_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_030_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_030_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_031_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_031_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_032_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_032_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_033_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_033_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_035_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_035_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_036_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_036_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_037_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_037_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_038_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_038_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_041_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_041_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_044_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_044_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_045_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_045_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_046_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_046_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_050_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_050_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_058_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_058_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_060_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_060_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_062_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_062_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_063_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_063_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_065_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_065_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_066_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_066_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_067_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_067_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_068_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_068_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_070_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_070_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_072_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_072_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_075_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_075_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_079_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_016_azCenter_079_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_011_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_011_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_014_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_014_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_015_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_015_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_016_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_016_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_018_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_018_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_019_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_019_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_020_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_020_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_021_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_021_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_023_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_023_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_024_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_024_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_025_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_025_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_026_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_026_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_029_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_029_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_030_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_030_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_032_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_032_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_034_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_034_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_035_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_035_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_037_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_037_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_038_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_038_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_039_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_039_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_040_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_040_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_043_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_043_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_044_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_044_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_045_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_045_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_046_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_046_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_047_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_047_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_048_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_048_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_049_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_049_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_050_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_050_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_051_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_051_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_052_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_052_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_053_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_053_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_054_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_054_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_056_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_056_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_057_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_057_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_058_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_058_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_061_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_061_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_062_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_062_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_063_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_063_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_064_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_064_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_066_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_066_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_067_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_067_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_068_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_068_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_069_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_069_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_071_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_071_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_072_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_072_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_074_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_074_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_075_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_075_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_079_00_serial_c71.mat",
            "btr70_synth_A_elevDeg_017_azCenter_079_00_serial_c71.mat",
        ],
        "m1": [
            "m1_synth_A_elevDeg_014_azCenter_010_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_010_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_011_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_011_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_012_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_012_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_014_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_014_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_015_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_015_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_016_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_016_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_017_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_017_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_018_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_018_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_019_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_019_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_020_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_020_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_021_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_021_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_022_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_022_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_024_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_024_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_026_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_026_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_028_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_028_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_029_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_029_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_030_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_030_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_031_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_031_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_032_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_032_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_033_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_033_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_034_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_034_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_036_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_036_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_037_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_037_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_038_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_038_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_039_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_039_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_040_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_014_azCenter_040_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_010_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_010_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_011_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_011_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_012_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_012_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_013_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_013_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_014_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_014_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_015_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_015_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_016_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_016_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_017_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_017_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_018_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_018_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_019_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_019_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_020_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_020_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_021_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_021_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_022_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_022_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_023_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_023_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_024_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_024_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_025_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_025_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_026_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_026_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_027_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_027_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_028_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_028_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_029_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_029_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_030_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_030_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_031_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_031_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_032_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_032_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_033_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_033_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_034_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_034_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_035_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_035_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_036_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_036_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_037_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_037_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_038_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_038_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_039_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_039_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_040_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_040_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_041_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_041_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_042_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_042_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_043_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_043_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_044_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_044_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_045_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_045_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_046_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_046_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_047_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_047_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_049_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_049_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_050_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_050_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_051_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_051_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_052_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_052_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_055_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_055_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_058_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_058_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_059_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_059_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_060_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_060_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_064_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_064_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_072_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_072_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_074_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_074_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_076_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_076_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_077_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_077_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_079_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_016_azCenter_079_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_012_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_012_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_013_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_013_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_014_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_014_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_015_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_015_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_016_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_016_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_017_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_017_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_018_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_018_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_019_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_019_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_020_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_020_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_021_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_021_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_024_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_024_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_025_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_025_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_028_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_028_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_029_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_029_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_030_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_030_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_032_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_032_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_033_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_033_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_034_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_034_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_035_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_035_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_037_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_037_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_038_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_038_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_039_18_serial_0ap00n.mat",
            "m1_synth_A_elevDeg_017_azCenter_039_18_serial_0ap00n.mat",
        ],
        "m2": [
            "m2_synth_A_elevDeg_014_azCenter_011_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_011_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_012_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_012_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_013_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_013_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_014_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_014_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_015_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_015_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_016_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_016_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_017_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_017_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_018_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_018_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_019_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_019_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_021_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_021_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_023_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_023_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_025_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_025_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_026_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_026_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_027_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_027_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_028_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_028_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_029_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_029_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_030_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_030_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_031_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_031_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_033_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_033_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_034_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_034_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_035_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_035_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_036_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_036_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_037_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_014_azCenter_037_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_010_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_010_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_011_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_011_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_012_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_012_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_013_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_013_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_014_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_014_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_015_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_015_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_016_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_016_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_017_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_017_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_018_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_018_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_019_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_019_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_020_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_020_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_021_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_021_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_022_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_022_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_023_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_023_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_024_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_024_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_025_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_025_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_026_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_026_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_027_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_027_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_028_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_028_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_029_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_029_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_030_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_030_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_031_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_031_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_032_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_032_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_033_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_033_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_034_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_034_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_035_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_035_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_036_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_036_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_037_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_037_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_038_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_038_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_039_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_039_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_040_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_040_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_041_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_041_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_042_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_042_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_043_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_043_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_044_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_044_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_046_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_046_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_047_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_047_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_048_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_048_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_049_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_049_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_052_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_052_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_055_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_055_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_056_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_056_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_057_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_057_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_061_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_061_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_069_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_069_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_071_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_071_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_073_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_073_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_074_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_074_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_076_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_076_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_077_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_077_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_078_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_078_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_079_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_016_azCenter_079_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_010_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_010_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_011_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_011_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_012_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_012_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_013_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_013_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_014_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_014_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_015_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_015_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_016_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_016_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_017_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_017_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_018_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_018_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_021_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_021_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_022_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_022_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_025_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_025_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_026_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_026_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_027_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_027_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_029_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_029_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_030_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_030_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_031_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_031_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_032_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_032_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_034_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_034_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_035_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_035_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_036_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_036_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_037_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_037_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_040_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_040_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_041_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_041_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_043_91_serial_mv02gx.mat",
            "m2_synth_A_elevDeg_017_azCenter_043_91_serial_mv02gx.mat",
        ],
        "m35": [
            "m35_synth_A_elevDeg_014_azCenter_010_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_010_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_012_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_012_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_013_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_013_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_014_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_014_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_015_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_015_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_016_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_016_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_017_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_017_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_018_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_018_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_019_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_019_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_020_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_020_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_022_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_022_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_024_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_024_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_026_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_026_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_027_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_027_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_028_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_028_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_029_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_029_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_030_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_030_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_031_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_031_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_032_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_032_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_034_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_034_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_035_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_035_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_036_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_036_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_037_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_037_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_038_62_serial_t839.mat",
            "m35_synth_A_elevDeg_014_azCenter_038_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_010_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_010_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_011_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_011_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_012_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_012_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_013_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_013_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_014_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_014_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_015_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_015_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_016_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_016_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_017_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_017_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_018_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_018_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_019_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_019_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_020_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_020_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_021_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_021_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_022_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_022_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_023_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_023_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_024_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_024_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_025_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_025_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_026_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_026_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_027_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_027_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_028_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_028_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_029_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_029_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_030_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_030_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_031_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_031_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_032_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_032_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_033_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_033_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_034_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_034_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_035_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_035_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_036_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_036_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_037_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_037_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_038_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_038_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_039_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_039_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_040_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_040_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_041_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_041_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_042_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_042_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_043_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_043_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_044_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_044_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_045_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_045_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_047_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_047_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_048_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_048_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_049_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_049_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_050_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_050_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_053_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_053_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_056_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_056_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_057_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_057_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_058_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_058_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_062_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_062_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_070_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_070_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_072_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_072_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_074_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_074_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_075_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_075_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_077_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_077_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_078_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_078_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_079_62_serial_t839.mat",
            "m35_synth_A_elevDeg_016_azCenter_079_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_010_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_010_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_011_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_011_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_012_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_012_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_013_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_013_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_014_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_014_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_015_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_015_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_016_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_016_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_017_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_017_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_018_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_018_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_019_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_019_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_022_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_022_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_023_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_023_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_026_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_026_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_027_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_027_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_028_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_028_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_030_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_030_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_031_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_031_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_032_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_032_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_033_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_033_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_035_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_035_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_036_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_036_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_037_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_037_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_038_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_038_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_041_62_serial_t839.mat",
            "m35_synth_A_elevDeg_017_azCenter_041_62_serial_t839.mat",
        ],
        "m548": [
            "m548_synth_A_elevDeg_014_azCenter_011_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_011_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_012_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_012_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_013_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_013_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_014_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_014_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_015_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_015_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_016_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_016_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_017_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_017_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_018_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_018_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_019_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_019_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_021_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_021_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_023_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_023_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_025_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_025_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_026_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_026_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_027_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_027_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_028_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_028_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_029_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_029_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_030_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_030_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_031_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_031_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_033_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_033_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_034_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_034_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_035_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_035_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_036_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_036_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_037_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_014_azCenter_037_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_010_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_010_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_011_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_011_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_012_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_012_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_013_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_013_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_014_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_014_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_015_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_015_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_016_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_016_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_017_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_017_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_018_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_018_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_019_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_019_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_020_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_020_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_021_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_021_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_022_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_022_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_023_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_023_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_024_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_024_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_025_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_025_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_026_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_026_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_027_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_027_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_028_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_028_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_029_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_029_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_030_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_030_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_031_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_031_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_032_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_032_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_033_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_033_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_034_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_034_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_035_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_035_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_036_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_036_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_037_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_037_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_038_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_038_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_039_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_039_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_040_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_040_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_041_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_041_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_042_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_042_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_043_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_043_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_044_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_044_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_046_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_046_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_047_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_047_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_048_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_048_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_049_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_049_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_052_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_052_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_055_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_055_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_056_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_056_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_057_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_057_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_061_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_061_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_069_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_069_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_071_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_071_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_073_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_073_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_074_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_074_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_076_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_076_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_077_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_077_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_078_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_078_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_079_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_016_azCenter_079_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_010_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_010_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_011_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_011_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_012_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_012_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_013_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_013_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_014_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_014_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_015_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_015_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_016_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_016_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_017_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_017_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_018_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_018_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_021_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_021_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_022_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_022_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_025_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_025_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_026_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_026_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_027_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_027_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_029_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_029_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_030_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_030_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_031_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_031_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_032_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_032_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_034_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_034_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_035_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_035_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_036_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_036_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_037_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_037_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_040_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_040_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_041_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_041_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_043_63_serial_c245hab.mat",
            "m548_synth_A_elevDeg_017_azCenter_043_63_serial_c245hab.mat",
        ],
        "m60": [
            "m60_synth_A_elevDeg_015_azCenter_010_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_010_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_011_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_011_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_012_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_012_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_013_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_013_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_014_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_014_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_015_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_015_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_016_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_016_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_017_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_017_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_018_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_018_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_019_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_019_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_020_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_020_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_021_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_021_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_022_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_022_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_023_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_023_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_024_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_024_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_025_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_025_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_026_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_026_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_027_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_027_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_028_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_028_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_029_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_029_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_030_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_030_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_031_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_031_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_032_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_032_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_033_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_033_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_034_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_034_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_035_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_035_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_036_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_036_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_037_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_037_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_038_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_038_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_039_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_039_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_040_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_040_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_042_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_042_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_043_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_043_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_044_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_044_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_045_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_045_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_046_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_046_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_047_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_047_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_048_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_048_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_049_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_049_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_050_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_050_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_052_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_052_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_053_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_053_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_054_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_054_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_055_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_055_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_056_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_056_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_057_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_057_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_058_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_058_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_059_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_059_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_060_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_060_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_061_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_061_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_062_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_062_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_063_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_063_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_064_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_064_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_065_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_065_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_066_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_066_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_067_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_067_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_068_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_068_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_070_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_070_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_073_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_073_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_074_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_074_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_075_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_075_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_076_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_076_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_077_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_077_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_078_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_078_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_079_74_serial_3336.mat",
            "m60_synth_A_elevDeg_015_azCenter_079_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_010_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_010_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_011_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_011_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_012_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_012_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_013_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_013_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_014_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_014_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_015_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_015_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_016_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_016_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_017_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_017_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_018_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_018_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_019_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_019_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_020_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_020_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_021_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_021_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_022_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_022_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_023_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_023_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_024_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_024_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_025_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_025_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_027_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_027_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_028_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_028_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_029_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_029_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_030_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_030_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_031_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_031_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_033_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_033_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_034_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_034_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_035_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_035_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_036_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_036_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_037_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_037_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_038_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_038_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_039_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_039_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_040_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_040_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_041_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_041_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_042_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_042_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_043_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_043_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_045_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_045_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_046_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_046_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_048_74_serial_3336.mat",
            "m60_synth_A_elevDeg_016_azCenter_048_74_serial_3336.mat",
        ],
        "t72": [
            "t72_synth_A_elevDeg_016_azCenter_013_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_013_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_015_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_015_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_016_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_016_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_017_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_017_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_020_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_020_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_021_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_021_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_022_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_022_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_027_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_027_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_028_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_028_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_029_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_029_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_030_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_030_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_032_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_032_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_033_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_033_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_034_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_034_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_035_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_035_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_036_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_036_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_037_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_037_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_038_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_038_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_039_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_039_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_040_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_040_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_041_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_041_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_042_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_042_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_043_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_043_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_044_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_044_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_046_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_046_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_047_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_047_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_048_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_048_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_049_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_049_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_050_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_050_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_051_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_051_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_052_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_052_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_053_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_053_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_054_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_054_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_055_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_055_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_056_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_056_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_057_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_057_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_058_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_058_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_059_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_059_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_060_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_060_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_061_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_061_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_062_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_062_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_063_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_063_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_064_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_064_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_065_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_065_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_066_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_066_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_068_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_068_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_069_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_069_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_070_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_070_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_071_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_071_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_072_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_072_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_073_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_073_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_074_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_074_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_075_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_075_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_076_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_076_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_078_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_078_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_079_77_serial_812.mat",
            "t72_synth_A_elevDeg_016_azCenter_079_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_011_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_011_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_012_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_012_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_013_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_013_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_014_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_014_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_016_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_016_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_017_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_017_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_018_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_018_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_019_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_019_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_022_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_022_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_023_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_023_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_024_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_024_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_025_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_025_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_026_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_026_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_027_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_027_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_028_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_028_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_029_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_029_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_031_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_031_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_032_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_032_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_033_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_033_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_034_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_034_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_035_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_035_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_036_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_036_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_037_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_037_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_038_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_038_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_041_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_041_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_042_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_042_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_043_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_043_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_044_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_044_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_045_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_045_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_046_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_046_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_047_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_047_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_048_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_048_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_049_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_049_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_050_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_050_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_053_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_053_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_054_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_054_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_057_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_057_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_058_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_058_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_059_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_059_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_061_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_061_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_062_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_062_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_063_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_063_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_064_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_064_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_066_77_serial_812.mat",
            "t72_synth_A_elevDeg_017_azCenter_066_77_serial_812.mat",
        ],
        "zsu23": [
            "zsu23_synth_A_elevDeg_015_azCenter_010_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_010_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_011_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_011_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_012_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_012_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_013_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_013_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_014_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_014_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_015_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_015_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_016_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_016_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_017_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_017_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_018_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_018_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_019_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_019_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_020_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_020_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_021_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_021_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_022_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_022_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_023_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_023_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_024_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_024_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_025_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_025_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_026_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_026_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_027_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_027_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_028_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_028_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_029_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_029_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_030_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_030_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_031_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_031_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_032_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_032_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_033_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_033_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_034_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_034_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_035_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_035_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_036_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_036_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_037_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_037_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_039_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_039_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_040_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_040_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_041_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_041_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_042_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_042_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_043_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_043_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_044_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_044_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_045_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_045_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_046_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_046_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_047_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_047_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_049_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_049_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_050_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_050_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_051_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_051_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_052_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_052_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_053_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_053_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_054_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_054_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_055_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_055_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_056_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_056_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_057_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_057_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_058_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_058_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_059_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_059_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_060_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_060_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_061_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_061_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_062_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_062_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_063_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_063_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_064_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_064_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_065_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_065_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_067_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_067_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_069_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_069_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_070_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_070_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_071_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_071_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_072_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_072_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_073_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_073_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_074_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_074_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_075_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_075_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_076_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_076_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_077_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_077_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_078_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_078_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_079_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_015_azCenter_079_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_010_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_010_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_011_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_011_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_012_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_012_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_013_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_013_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_014_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_014_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_015_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_015_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_016_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_016_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_017_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_017_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_018_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_018_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_019_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_019_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_020_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_020_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_021_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_021_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_022_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_022_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_024_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_024_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_025_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_025_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_026_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_026_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_027_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_027_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_028_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_028_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_030_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_030_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_031_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_031_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_032_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_032_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_033_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_033_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_034_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_034_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_035_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_035_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_036_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_036_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_037_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_037_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_038_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_038_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_039_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_039_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_040_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_040_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_042_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_042_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_043_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_043_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_045_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_045_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_047_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_047_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_048_99_serial_d08.mat",
            "zsu23_synth_A_elevDeg_016_azCenter_048_99_serial_d08.mat",
        ],
    },
}
